# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import os
import sys
import click
import importlib

REQUIRED = [
    'CASSANDRA_SEEDS',
    'CASSANDRA_KEYSPACE'
]
OPTIONAL = [
    'CASSANDRA_PORT',
    'CASSANDRA_USER',
    'CASSANDRA_PASSWORD',
    'CASSANDRA_CQLVERSION'
]


def get_config(env_override=None):
    """ Get the configuration dict. """
    env = os.environ
    if env_override is not None:
        for key, value in env_override.iteritems():
            env[key] = value

    settings = None
    try:
        if 'CASSANDRA_SETTINGS' in os.environ:
            settings = importlib.import_module(env['CASSANDRA_SETTINGS'])
    except Exception:
        click.secho('Unable to load settings module {}!'.format(env.get('CASSANDRA_SETTINGS')), fg='red')
        sys.exit()

    # Check configuration
    config = {}

    for c in REQUIRED:
        if hasattr(settings, c):
            config[c.lower().split('_', 1).pop()] = getattr(settings, c)
        else:
            click.secho('{} is missing is settings file!'.format(c), fg='red')
            sys.exit()
    for c in OPTIONAL:
        if hasattr(settings, c):
            config[c.lower().split('_', 1).pop()] = getattr(settings, c)

    return config



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import os
import sys
import click
import time
import warnings

from .db import get_current_schema, get_session
from .db import update_snapshot

warnings.filterwarnings("ignore")


def get_last_migration(config):
    """
    Get the last migration stored on cassandra.
    If there is no first migration, it will return 0
    If there is no table shift_migrations, it will return None
    """
    get_session().set_keyspace(config['keyspace'])
    try:
        migrations = get_session().execute("SELECT migration FROM shift_migrations LIMIT 1")
        if not migrations:
            return 0
        last_migration = migrations[0].migration
        if last_migration.endswith('.cql'):
            last_migration = last_migration[:-4]
    except Exception as e:
        last_migration = None
    return last_migration


def get_migrations_on_file():
    """ Get the stored migrations on file. """
    try:
        files = []
        for f in os.listdir('migrations'):
            if f[-3:] == 'cql':
                files.append(f)
    except Exception:
        click.secho('Unable to open the migrations directory!', fg='red')
        sys.exit()
    files.sort()
    return files


def get_head_migration_on_file(migrations):
    """ Get the highest migration on file. """
    mig = []
    for m in migrations:
        try:
            m = m.split('_')[0]
            m = int(m)
        except Exception:
            continue
        mig.append(m)
    if not mig:
        return 0
    return mig.pop()


def get_pending_migrations(last_migration, migrations, head=None):
    """
    Return a list of migration files that should be applied.

    If head is not None, it must be a positive integer pointing at the tarjet migration we
    are headed. If the current DB migration is ahead of the head, then the delta migrations
    will be applied DOWN. Otherwise the delta migrations will be applied UP.

    In case DOWN migrations are not found in the file, the migration wont be able to continue.
    Pending migrations will be returned IN ORDER in which they must be executed.
    """
    if last_migration and '{}.cql'.format(last_migration) not in migrations:
        click.secho('Unable to migrate because migrations DB is ahead of migrations on file.', fg='red')
        sys.exit()
    if not last_migration:
        last_migration = 0
    else:
        last_migration = last_migration.split('_')[0]

    pointer = int(last_migration)
    files_head = get_head_migration_on_file(migrations)
    # If no head is specified, then the target head will be the
    # largest migration on file.
    target = files_head if head is None else int(head)
    if target > files_head:
        click.secho('The target migration provided does not exist in the migrations dir.', fg='red')
        sys.exit()
    # Are we migrating up or down?
    up = True if pointer <= target else False

    if not up:
        migrations.sort(reverse=True)

    pending = []
    for m in migrations:
        try:
            f = m.split('_')[0]
            if f == '00000':
                continue
            cur_mig = int(f)
        except Exception:
            continue

        if up:
            if cur_mig <= pointer or cur_mig > target:
                continue
        else:
            if cur_mig <= target or cur_mig > pointer:
                continue

        pending.append(m)
    return (pending, up)


def apply_migration(file, up, keyspace):
    """
    Apply the migration given the raw file name.
    If up is True, then it will execute the up statement, else it will execute the down statement.
    Valid CQL format in files is as follows:

    --UP--
    /* Your CQL statements here, separated by ; */
    --DOWN--
    /* Your CQL statements here. They MUST revert what the UP statements do */

    """
    fname = file
    click.echo("Applying migration {} {} ".format(file, ('UP' if up else 'DOWN')), nl=False)
    try:
        file = open('migrations/{}'.format(file), 'r')
        content = file.read()
        file.close()
    except Exception:
        click.secho('ERROR', fg='red', bold=True)
        return (False, 'Unable to open file {}.'.format(file))
    content = content.replace('--UP--', '', 1)
    parts = content.split('--DOWN--')
    if len(parts) > 1:
        qryup, qrydown = parts
    else:
        qryup = parts[0]
        qrydown = None

    if not qrydown:
        click.secho('ERROR', fg='red', bold=True)
        return (False, 'File {} does not include a --DOWN-- statement.'.format(fname))

    if keyspace is not None:
        get_session().set_keyspace(keyspace)
    qry = qryup if up else qrydown
    try:
        for q in qry.replace('\n', '').split(';'):
            q = q.strip()
            if q != '':
                get_session().execute(q.strip())
    except Exception as e:
        click.secho('ERROR', fg='red', bold=True)
        return (False, e)
    click.secho('OK', fg='green', bold=True)
    return (True, None)


def create_migration_file(name, up, down=None, title='', description='',
                          genesis=False):
    """
    Create a migration file in the migrations folder
    and return its filename.
    """
    if not os.path.isdir('migrations'):
        os.mkdir('migrations')
    migrations = get_migrations_on_file()
    i = 1
    while True:
        count = len(migrations) if not genesis else -1
        file_name = [str(count + i).zfill(5)]
        if not genesis:
            file_name.append(name.strip().lower().replace(' ', '_'))
        file_name = '_'.join(file_name) + '.cql'
        if os.path.isfile('migrations/' + file_name):
            i += 1
            continue
        file = open('migrations/' + file_name, 'w')
        file.write('/*\n')
        if title:
            file.write(title)
        else:
            file.write(name)
        file.write('\n\n')
        if description:
            file.write(description + '\n')
        file.write('Created: ' + time.strftime("%d-%m-%Y") + '\n')
        file.write('*/\n')
        file.write('--UP--\n')
        file.write(up + '\n\n')
        if down:
            file.write('--DOWN--\n')
            file.write(down)
        return file_name


def create_init_migration(config):
    """
    Create the genesis configuration file and return it's filename.
    This function will also write the current keyspace schema (the one
    defined in the current settings) as the content of the file.
    """
    click.echo("Creating migration genesis... ", nl=False)
    if os.path.isfile('migrations/00000.cql'):
        click.secho('ERROR (already exists)', fg='red', bold=True)
        return False
    current = get_current_schema(config)
    down = 'DROP KEYSPACE {};'.format(config.get('keyspace'))
    new_file = create_migration_file(name='', title='MIGRATION GENESIS', up=current, down=down, genesis=True)
    click.secho('OK', fg='green', bold=True)
    update_snapshot(current)
    return new_file



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

class Column(object):
    def __init__(self, name, type, kind='regular', order=None, position=-1):
        self.name = name
        self.type = type
        self.kind = kind
        self.order = order
        self.position = position

    def is_pk(self):
        return self.kind == 'partition_key'

    def is_clustering(self):
        return self.kind == 'clustering'

    def dump_cql(self):
        return '{} {}'.format(self.name, self.type.upper())

    def __eq__(self, other):
        if not isinstance(other, Column):
            return False
        return (self.name == other.name and
                self.type.lower() == other.type.lower() and
                self.kind == other.kind and
                self.order == other.order and
                self.position == other.position)

    def __ne__(self, other):
        return not self.__eq__(other)


class Table(object):
    def __init__(self, name, columns=[]):
        self.name = name
        self.columns = columns

    def primary_keys(self):
        pk = {}
        for c in self.columns:
            if c.is_pk():
                pk[c.position] = c.name
        r = []
        for i in sorted(pk):
            r.append(pk[i])
        return r

    def clustering_columns(self):
        cl = {}
        for c in self.columns:
            if c.is_clustering():
                cl[c.position] = c.name
        r = []
        for i in sorted(cl):
            r.append(cl[i])
        return r

    def get_column(self, name):
        for c in self.columns:
            if c.name == name:
                return c
        return None

    def dump_cql(self):
        cql = ['CREATE TABLE {} ('.format(self.name)]
        cql.append('\t' + ',\n\t'.join([x.dump_cql() for x in self.columns]) + ",")
        pk = self.primary_keys()
        pks = ('({})' if len(pk) > 1 else '{}').format(', '.join(pk))
        cc = self.clustering_columns()
        clustering = ', ' + ', '.join(cc) if cc else ''
        cql.append('\tPRIMARY KEY ({}{})'.format(pks, clustering))
        cql.append(')')
        return '\n'.join(cql)

    def __eq__(self, other):
        if not isinstance(other, Table):
            return False
        if len(self.columns) != len(other.columns):
            return False
        for col in self.columns:
            other_col = other.get_column(col.name)
            if not other_col:
                return False
            if col != other_col:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def drop_cql(self):
        return 'DROP TABLE "{}"'.format(self.name)

    def drop_column_cql(self, column):
        return 'ALTER TABLE "{}" DROP "{}"'.format(self.name, column.name)

    def add_column_cql(self, column):
        return 'ALTER TABLE "{}" ADD "{}" {}'.format(self.name, column.name, column.type)

    def alter_column_type_cql(self, name, type):
        return 'ALTER TABLE "{}" ALTER "{}" TYPE {}'.format(self.name, name, type)


class Keyspace(object):
    def __init__(self, name, tables=[]):
        self.name = name
        self.tables = tables

    def get_table(self, name):
        for t in self.tables:
            if t.name == name:
                return t
        return None


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import click
import warnings

from .migrate import create_migration_file, create_init_migration, get_migrations_on_file
from .migrate import get_last_migration, get_pending_migrations, apply_migration
from .db import connect, get_current_schema, create_demo_keyspace, keyspace_exists
from .db import record_migration, delete_demo_keyspace, create_migration_table, DEMO_KEYSPACE
from .db import auto_migrate_keyspace, get_snapshot, update_snapshot
from .config import get_config

warnings.filterwarnings("ignore")

# Get configuration
config = get_config()

@click.group()
def cli():
    pass


@cli.command('create', short_help='Create a new migration file.')
@click.argument('name', required=True)
@click.option('--title', help='Migration title', default=None)
@click.option('--description', help='Migration description', default=None)
def create(name, title, description):
    """ Create a new migration file. """
    file = create_migration_file(name=name, up='/* YOUR CQL GOES HERE */', 
                                 title=title, description=description)
    click.echo('Created migration file ', nl=False)
    click.secho(file, bold=True, fg='green')


@cli.command('init', short_help='Create the migration genesis based on the current keyspace.')
def init():
    """ Initiate the migration project in the current directory. """
    global config
    # Cassandra connection.
    connect(config)
    create_init_migration(config)


@cli.command('status', short_help='Show the current migration status.')
@click.option('--settings', default=None, help='Settings module (not file). Must contain CASSANDRA_SEEDS and CASSANDRA_KEYSPACE defined')
def status(settings):
    """ Get the current migration status. """
    global config
    if settings is not None:
        config = get_config({'CASSANDRA_SETTINGS': settings})
    # Cassandra connection.
    connect(config)
    # Check migrations on file.
    migrations = get_migrations_on_file()
    if not migrations:
        click.secho('No migrations found on the current directory.')
        return
    if '00000.cql' not in migrations:
        click.secho('There is no genesis migration found.')
        return
    if not keyspace_exists(config.get('keyspace')):
        click.echo('Shift hasn\'t been initialized on this keyspace.\nRun \'shift migrate\' to initiate or user the --help flag.')
        return
    last = get_last_migration(config)
    if last is None:
        click.secho('Shift hasn\'t been initialized in this keyspace.')
        return
    pending, up = get_pending_migrations(last, migrations)
    if len(pending) <= 0:
        click.echo("Already up to date.\nCurrent head is {}".format(last))
        return
    click.echo("Cassandra is {} movements behind the current file head ({}).\nCurrent Cassandra head is {}".format(len(pending), migrations[-1], last))


@cli.command('auto-update', short_help='Auto generate the next migration targeting the current Cassandra structure.')
@click.option('--print', is_flag=True, help='Just print the migrations')
@click.option('--name', required=True, help='Name of the update')
def auto_update(print, name):
    global config
    # Cassandra connection.
    connect(config)
    # Check migrations on file.
    migrations = get_migrations_on_file()
    if not keyspace_exists(config.get('keyspace')):
        click.echo('Shift hasn\'t been initialized on this keyspace.\nRun \'shift migrate\' to initiate or user the --help flag.')
        return
    last = get_last_migration(config)
    if last is None:
        click.secho('Shift hasn\'t been initialized in this keyspace.')
        return
    pending, up = get_pending_migrations(last, migrations)
    if len(pending) > 0:
        click.secho('There are pending migrations to be done, please migrate before auto-updating.')
        return
    # Create demo keyspace to compare
    snap = get_snapshot()
    if not snap:
        click.secho('Unable to locate the last snapshot.', fg='red')
        return
    create_demo_keyspace(snap, config.get('keyspace'))
    actions_up = auto_migrate_keyspace(DEMO_KEYSPACE, config.get('keyspace'))
    if len(actions_up) <= 0:
        click.secho('Cassandra is up to date with migrations on file.')
        delete_demo_keyspace()
        return
    actions_down = auto_migrate_keyspace(config.get('keyspace'), DEMO_KEYSPACE)
    delete_demo_keyspace()
    upquery = ';\n'.join(actions_up) + ";"
    downquery = ';\n'.join(actions_down)  + ";"
    if print:
        click.echo('---\n' + upquery + '\n---\n')
        return
    file = create_migration_file(name, upquery, downquery)
    click.echo('Created migration file ', nl=False)
    record_migration(file, get_current_schema(config), config)
    click.secho(file, bold=True, fg='green')


@cli.command('migrate', short_help='Migrate the current database.')
@click.argument('head', required=False)
@click.option('--simulate', is_flag=True, help='Just print the migrations that will be performed')
@click.option('--just-demo', is_flag=True, help='Just perform the migrations in demo DB')
@click.option('--settings', default=None, help='Settings module (not file). Must contain CASSANDRA_SEEDS and CASSANDRA_KEYSPACE defined')
def migrate(head, simulate, just_demo, settings):
    """ Migrate now. """
    global config
    if settings is not None:
        config = get_config({'CASSANDRA_SETTINGS': settings})
    # Input validation.
    try:
        head = int(head) if head else None
    except Exception:
        click.secho('Head argument must be an integer.', fg='red')
        return
    # Cassandra connection.
    connect(config)
    # Check migrations on file.
    migrations = get_migrations_on_file()
    # Check if there is the migration genesis is present
    # if it's not present we cannot continue.
    if '00000.cql' not in migrations:
        click.secho('Migration genesis (00000.cql) is missing! Forgot to run init command first?', fg='red')
        return
    # Check if the keyspace exists and if we have a migrations
    # table configured.
    if not keyspace_exists(config.get('keyspace')):
        # Keyspace does not exist, we need to create it based on the genesis file.
        click.echo('Keyspace not found, creating from the genesis file.')
        result, err = apply_migration('00000.cql', True, None)
        if not result:
            click.secho('---\nUnable to continue due to an error genesis migration:\n\n{}\n---\n'.format(err.message), fg='red')
            return
        # Override head, it needs to go all the way from the bottom...
        head = None

    schema = get_current_schema(config)
    last = get_last_migration(config)
    if last is None:
        result, err = create_migration_table(config.get('keyspace'))
        if not result:
            click.secho('---\nUnable to continue due to an error:\n\n{}\n---\n'.format(err.message), fg='red')
            return
        update_snapshot(get_current_schema(config))

    if len(migrations) <= 0:
        create_init_migration(config)
    pending, up = get_pending_migrations(last, migrations, head)
    if len(pending) <= 0:
        click.echo("Already up to date.")
        return

    if simulate:
        for p in pending:
            click.echo('{} will be applied {}'.format(p, 'UP' if up else 'DOWN'))
        return

    # First in demo
    error = False
    create_demo_keyspace(schema, config['keyspace'])
    for f in pending:
        res, err = apply_migration(file=f, up=up, keyspace=DEMO_KEYSPACE)
        if not res:
            error = True
            click.secho('---\nUnable to continue due to an error in {}:\n\n{}\n---\n'.format(f, err), fg='red')
            break
    delete_demo_keyspace()
    if error:
        return
    if just_demo:
        return
    # Now in real keyspace
    for f in pending:
        res, err = apply_migration(file=f, up=up, keyspace=config['keyspace'])
        if not res:
            error = True
            click.secho('---\nUnable to continue due to an error in {}:\n\n{}\n---\n'.format(f, err.message), fg='red')
            break
        record_migration(name=f, schema=get_current_schema(config), up=up, config=config)
    if error:
        return
    click.echo("Migration completed successfully.")


if __name__ == "__main__":
    cli()



from setuptools import setup

setup(
    name='cql-shifter',
    description='A tool for migration management with Cassandra',
    long_description='shifter gets the pain away from managing migrations with Cassandra',
    url='http://github.com/rgamba/shift/',
    author='Ricardo Gamba',
    author_email='rgamba@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
    version='0.2',
    keywords='development database migration cassandra',
    include_package_data=True,
    packages=['shifter'],
    install_requires=[
        'click>=6.6',
        'cassandra-driver>=3.7.0',
        'futures>=3.0.5',
        'invoke>=0.13.0',
        'six>=1.10.0'
        'cqlsh'
    ],
    entry_points={
        'console_scripts': [
            'shifter = shifter.cli:cli',
        ],
    },
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import sys
import time
import hashlib
import copy

import click
from invoke import run
from cassandra.cluster import Cluster
from cassandra.util import max_uuid_from_time
from cassandra.auth import PlainTextAuthProvider

from .map import Column, Table, Keyspace


DEMO_KEYSPACE = 'cm_tmp'
SHIFT_TABLE = 'shift_migrations'

session = None


def connect(config):
    global session
    # Connect to cassandra
    auth_provider = None
    if config.get('user'):
        auth_provider = PlainTextAuthProvider(username=config.get('user'), password=config.get('password'))
    cluster = Cluster(
        contact_points=config.get('seeds'),
        port=(int(config.get('port')) if config.get('port') else 9042),
        auth_provider=auth_provider
    )
    try:
        session = cluster.connect()
    except:
        click.secho("Unable to connect to Cassandra", fg='red')
        sys.exit()
    return session


def get_session():
    global session
    if session is None:
        connect()
    return session


def get_snapshot():
    try:
        file = open('migrations/.snapshot', 'r')
        content = file.read()
        file.close()
    except Exception:
        return None
    return content


def update_snapshot(schema):
    try:
        file = open('migrations/.snapshot', 'w+')
        file.write(schema)
        file.close()
    except Exception:
        pass


def run_cqlsh(config, command, keyspace=None):
    q = ['cqlsh', '-e', '"{}"'.format(command)]
    if config.get('user'):
        q.append('-u')
        q.append(config.get('user'))
    if config.get('password'):
        q.append('-p')
        q.append(config.get('password'))
    if keyspace:
        q.append('-k')
        q.append(keyspace)
    if config.get('cqlversion'):
        q.append('--cqlversion={}'.format(config.get('cqlversion')))
    q.append(config['seeds'][0])
    if config.get('port'):
        q.append(config['port'])
    return ' '.join(q)


def get_current_schema(config):
    try:
        cqlsh = run_cqlsh(config, command="DESCRIBE " + config['keyspace'])
        out = run(cqlsh, hide='stdout')
    except Exception as e:
        click.secho("Unable to get the current DB schema: {}".format(e), fg='red')
        sys.exit()
    return out.stdout


def create_demo_keyspace(schema, schema_name):
    schema = schema.replace("CREATE KEYSPACE {}".format(schema_name), "CREATE KEYSPACE {}".format(DEMO_KEYSPACE), 1)
    schema = schema.replace("{}.".format(schema_name), "{}.".format(DEMO_KEYSPACE))
    try:
        click.echo("Creating tmp keyspace... ", nl=False)
        session.execute("DROP KEYSPACE IF EXISTS {}".format(DEMO_KEYSPACE))
        for q in schema.replace('\n', '').split(';'):
            if q.strip() == "":
                continue
            session.execute(q)
    except Exception as e:
        click.secho("ERROR {}".format(e), fg='red', bold=True)
        sys.exit()
    click.secho("OK", fg='green', bold=True)


def delete_demo_keyspace():
    try:
        click.echo("Deleting tmp keyspace... ", nl=False)
        session.execute("DROP KEYSPACE IF EXISTS {}".format(DEMO_KEYSPACE))
        click.secho("OK", fg='green', bold=True)
    except Exception:
        click.secho("ERROR", fg='red', bold=True)


def record_migration(name, schema, config, up=True):
    session.set_keyspace(config['keyspace'])
    if name.endswith('.cql'):
        name = name[:-4]
    if not up:
        # Delete
        rows = session.execute(
            """
            SELECT time FROM shift_migrations 
            WHERE type = 'MIGRATION' 
                AND migration = %s 
            ALLOW FILTERING
            """, (name,)
        )
        if not rows:
            click.secho("Unable to select last migration from DB", fg="red")
            return False
        id = rows[0].time
        delete = session.execute("DELETE FROM shift_migrations WHERE type = 'MIGRATION' AND time = %s", (id,))
        return

    m = hashlib.md5()
    m.update(schema)
    session.execute(
        """
        INSERT INTO shift_migrations(type, time, migration, hash)
        VALUES (%s, %s, %s, %s)
        """,
        ('MIGRATION', max_uuid_from_time(time.time()), name, m.hexdigest())
    )
    update_snapshot(schema)


def create_migration_table(keyspace):
    session.set_keyspace(keyspace)
    click.echo("Creating shift_migrations table... ", nl=False)
    try:
        session.execute(
            """
            CREATE TABLE IF NOT EXISTS shift_migrations(
                type text,
                time timeuuid,
                migration text,
                hash text,
                PRIMARY KEY (type, time)
            )
            WITH CLUSTERING ORDER BY(time DESC)
            """
        )
        click.secho('OK', fg='green', bold=True)
        return (True, None)
    except Exception as e:
        click.secho('ERROR', fg='red', bold=True)
        return (False, e)


def keyspace_exists(name):
    session.set_keyspace('system_schema')
    ks = session.execute('SELECT keyspace_name FROM keyspaces')
    if not ks:
        return False
    for row in ks:
        if row.keyspace_name == name:
            return True
    return False


def auto_migrate_keyspace(source_name, target_name):
    """
    Compare the 2 keyspaces and return a list of
    queries to be performed in order to sync source to target.
    """
    source = Keyspace(name=source_name, tables=[])
    stables = get_keyspace_tables(source_name)
    for table in stables:
        source.tables.append(Table(
            name=table,
            columns=get_table_columns(source_name, table)
        ))
    target = Keyspace(name=target_name, tables=[])
    ttables = get_keyspace_tables(target_name)
    table = []
    for table in ttables:
        target.tables.append(Table(
            name=table,
            columns=get_table_columns(target_name, table)
        ))
    return get_keyspace_diff(source, target)


def get_keyspace_tables(keyspace):
    session.set_keyspace('system_schema')
    ks = session.execute('SELECT table_name FROM tables WHERE keyspace_name = %s', [keyspace])
    if not ks:
        return []
    tables = []
    for row in ks:
        tables.append(row.table_name)
    return tables


def get_table_columns(keyspace, table):
    """ Return a list of tuples (col_name, col_type). """
    session.set_keyspace('system_schema')
    ks = session.execute('SELECT * FROM columns WHERE keyspace_name = %s AND table_name = %s', [keyspace, table])
    if not ks:
        return []
    cols = []
    for row in ks:
        cols.append(Column(
            name=row.column_name,
            type=row.type,
            kind=row.kind,
            order=row.clustering_order,
            position=row.position))
    return cols


def get_columns_diff(source, target):
    if not isinstance(source, Column) or not isinstance(target, Column):
        raise ValueError("parameters must be Column instances")
    if source.kind != target.kind:
        return "kind"
    if source.position != target.position:
        return "position"
    if source.type != target.type:
        return "type"
    return None

def get_tables_diff(source, target):
    if not isinstance(source, Table) or not isinstance(target, Table):
        raise ValueError("parameters must be Table instances")
    if source == target:
        return None
    actions = []
    # Compare table name
    if source.name != target.name:
        raise ValueError("tables does not share the same name")
    # Columns in source table
    for col in source.columns:
        target_col = target.get_column(col.name)
        if not target_col:
            actions.append(source.drop_column_cql(col))
        else:
            if target_col != col:
                dif = get_columns_diff(col, target_col)
                if dif == "type":
                    actions.append(source.alter_column_type_cql(col.name, target_col.type))
    # Columns in destination table
    for col in target.columns:
        source_col = source.get_column(col.name)
        if not source_col:
            actions.append(source.add_column_cql(col))
    return actions


def get_keyspace_diff(source, target):
    if not isinstance(source, Keyspace) or not isinstance(target, Keyspace):
        raise ValueError("parameters must be Table instances")
    actions = []
    # Source tables
    for table in source.tables:
        target_table = target.get_table(table.name)
        if not target_table:
            actions.append(table.drop_cql())
        else:
            a = get_tables_diff(table, target_table)
            if a:
                actions += a
    # Target tables
    for table in target.tables:
        source_table = source.get_table(table.name)
        if not source_table:
            actions.append(table.dump_cql())
    return actions

